using System;
using System.Net.Sockets;
using System.Net;
using System.Windows.Forms;
using System.Text;
using System.IO;
using System.Collections;

using Phoenix;

namespace SpringClient
{
	/// <summary>
	/// Summary description for ConnectionManager.
	/// </summary>
	public class ConnectionManager
	{
		private Socket			s;
		private ClientMain		PrimaryForm;

		private SocketHandler	m_SocketHandler;

		private Queue			m_Chatids;

		public bool Connected
		{
			get
			{
				return s.Connected;
			}
		}

		public SocketHandler	socketHandler
		{
			get
			{
				return m_SocketHandler;
			}
		}

		public Socket socket
		{
			get
			{
				return s;
			}
		}

		public ConnectionManager()
		{
			//
			// TODO: Add constructor logic here
			//
		}

		public ConnectionManager(Socket sock)
		{
			s = sock;
			m_SocketHandler = new SocketHandler(s);
		}

		public bool Connect(string IP, int port)
		{
				s = new Socket(AddressFamily.InterNetwork, SocketType.Stream,
				ProtocolType.Tcp);

				IPEndPoint EPhost = null;

				IPHostEntry HostInfo = Dns.Resolve(Global.ServerLoc);
				IPAddress hostadd = HostInfo.AddressList[0];//IPAddress.Parse(IP);
				EPhost = new IPEndPoint(hostadd, port);
				
				s.Connect(EPhost);
				m_SocketHandler = new SocketHandler(s);

				m_Chatids = new Queue();

				return s.Connected;
		}

		public bool Connect(long IP, int port)
		{
				s = new Socket(AddressFamily.InterNetwork, SocketType.Stream,
					ProtocolType.Tcp);

				IPEndPoint EPhost = null;

				IPAddress hostadd = new IPAddress(Convert.ToInt64(IP));
				EPhost = new IPEndPoint(hostadd, port);

	//			MessageBox.Show(string.Format("connecting to {0} using port {1}", hostadd.ToString(), port.ToString()));
			
				s.Connect(EPhost);
				m_SocketHandler = new SocketHandler(s);

				m_Chatids = new Queue();

				return s.Connected;
		}

		public bool StartListener(ClientMain form)
		{
			try
			{
				if ( !s.Connected )
					return false;
				
				PrimaryForm = form;
				// asynchronous reading of 1 byte message header
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch
			{
				return false;
			}
			return true;
		}

		private void MessageHandler(byte[] msg)
		{
			System.Diagnostics.Debug.WriteLine(string.Format("Received msg: {0}", ((PACKETS)msg[0]).ToString()));
			try
			{
				//MessageBox.Show(((PACKETS)(msg[0])).ToString());
				switch ( (PACKETS)msg[0] )
				{
					case PACKETS.PACK_CHATLIST:
						m_SocketHandler.Read(2, new SocketHandler.delRead(HandlePackChatList), PrimaryForm);
						break;
					case PACKETS.PACK_JOINCHAT:
						m_SocketHandler.Read(20, new SocketHandler.delRead(HandlePackJoinChat), PrimaryForm);
						break;
					case PACKETS.PACK_CHATMESSAGE:
						m_SocketHandler.Read(2, new SocketHandler.delRead(HandlePackChatMessage), PrimaryForm);
						break;
					case PACKETS.PACK_USERENTERCHAT:
						m_SocketHandler.Read(20+CONSTANTS.MAXNAME+CONSTANTS.MAXCLAN, new SocketHandler.delRead(HandlePackUserEnterChat), PrimaryForm);
						break;
					case PACKETS.PACK_USERLEAVECHAT:
						m_SocketHandler.Read(20, new SocketHandler.delRead(HandlePackUserLeaveChat), PrimaryForm);
						break;						
					case PACKETS.PACK_BATTLELIST:
						m_SocketHandler.Read(4, new SocketHandler.delRead(HandlePackBattleList), PrimaryForm);
						break;
					case PACKETS.PACK_CREATECHAT:
						m_SocketHandler.Read(16+CONSTANTS.MAXCHATNAME, new SocketHandler.delRead(HandlePackCreateChat), PrimaryForm);
						break;
					case PACKETS.PACK_FAILEDCREATECHAT:
						HandlePackFailedCreateChat();
						break;
					case PACKETS.PACK_CHATDESTROYED:
						m_SocketHandler.Read(16, new SocketHandler.delRead(HandlePackChatDestroyed), PrimaryForm);
						break;
					case PACKETS.PACK_BATTLECREATESUCCESS:
						m_SocketHandler.Read(8+CONSTANTS.MAXMAPNAME, new SocketHandler.delRead(HandlePackBattleCreateSuccess), PrimaryForm);
						break;
					case PACKETS.PACK_BATTLECREATEFAILED:
						HandlePackBattleCreateFailed();
						break;
					case PACKETS.PACK_CREATEBATTLE:
						m_SocketHandler.Read(12+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME+CONSTANTS.MAXMAPNAME, new SocketHandler.delRead(HandlePackCreateBattle), Global.PrimaryForm);
						break;
					case PACKETS.PACK_DESTROYBATTLE:
						m_SocketHandler.Read(4, new SocketHandler.delRead(HandlePackDestroyBattle), PrimaryForm);
						break; 
					case PACKETS.PACK_BOOTFROMCHAT:
						m_SocketHandler.Read(4, new SocketHandler.delRead(HandlePackBootFromChat), PrimaryForm);
						break;
					case PACKETS.PACK_CHATSIZEUPDATE:
						m_SocketHandler.Read(18, new SocketHandler.delRead(HandlePackChatSizeUpdate), PrimaryForm);
						break;
					case PACKETS.PACK_BATTLESIZEUPDATE:
						m_SocketHandler.Read(5, new SocketHandler.delRead(HandlePackBattleSizeUpdate), PrimaryForm);
						break;
					case PACKETS.PACK_MAPUPDATE:
						m_SocketHandler.Read(CONSTANTS.MAXMAPNAME+4, new SocketHandler.delRead(HandlePackMapUpdate), PrimaryForm);
						break;
					case PACKETS.PACK_BATTLEPROGRESS:
						m_SocketHandler.Read(5, new SocketHandler.delRead(HandlePackBattleProgress), PrimaryForm);
						break;
					default:
						MessageBox.Show("Bad message");
						break;
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackBattleSizeUpdate(byte[] buffer)
		{
			try
			{
				uint id = ByteConverter.ToUInt32(buffer, 0);
				byte newsize = buffer[4];

				PrimaryForm.BattleSizeUpdate(id, newsize);

				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackMapUpdate(byte[] buffer )
		{
			try
			{
				uint id = ByteConverter.ToUInt32(buffer, 0);
				string newmap = Encoding.ASCII.GetString(buffer, 4, CONSTANTS.MAXMAPNAME);

				PrimaryForm.MapUpdate(id, newmap);
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackChatSizeUpdate(byte[] buffer)
		{
			try
			{
				int index = 0;
				Guid chatroom = ByteConverter.ToGUID(buffer, ref index);
				short newsize = ByteConverter.ToInt16(buffer, ref index);
				PrimaryForm.UpdateChatSize(chatroom, newsize);
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}
		private void HandlePackChatList(byte[] buffer)
		{
			try
			{
				short numlists = ByteConverter.ToInt16(buffer, 0);
				int messageLength = 18+CONSTANTS.MAXCHATNAME;

				for ( int i = 0; i < numlists; i++ )
				{
					m_SocketHandler.Read(messageLength, new SocketHandler.delRead(HandlePackChatListItem), PrimaryForm);
				}
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackChatListItem(byte[] buffer)
		{
			try
			{
				int index = 0;
				Guid id = ByteConverter.ToGUID(buffer, ref index);
				short size = ByteConverter.ToInt16(buffer, ref index);
				string name = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXCHATNAME);
				PrimaryForm.AddChat(name, id, size);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackJoinChat(byte[] buffer)
		{
			//MessageBox.Show("Enter Chat");
			int index = 0;
			Guid id = ByteConverter.ToGUID(buffer, ref index);
			int users = ByteConverter.ToInt32(buffer, ref index);

			PrimaryForm.EnterChat(id);
			PrimaryForm.UserChatAdd(id, UserInfo.ID, UserInfo.Name, UserInfo.Clan, false);
			int messagesize = 4+CONSTANTS.MAXNAME+CONSTANTS.MAXCLAN;
			messagesize *= users;
			m_Chatids.Enqueue(id);
			// and now each user individually
			m_SocketHandler.Read(messagesize, new SocketHandler.delRead(HandlePackJoinChatItem), PrimaryForm);
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackJoinChatItem(byte[] buffer)
		{
			try
			{
				Guid chatid = (Guid)m_Chatids.Dequeue();
				int offset = 0;
				while ( offset < buffer.Length )
				{
					int userid = ByteConverter.ToInt32(buffer, offset);
					offset += 4;
					string name = Encoding.ASCII.GetString(buffer, offset, CONSTANTS.MAXNAME).TrimEnd("\0".ToCharArray());
					offset += CONSTANTS.MAXNAME;
					string clan = Encoding.ASCII.GetString(buffer, offset, CONSTANTS.MAXCLAN).TrimEnd("\0".ToCharArray());
					offset += CONSTANTS.MAXCLAN;
					PrimaryForm.UserChatAdd(chatid, userid, name, clan, true);
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackChatMessage(byte[] buffer)
		{
			try
			{
				ushort len = ByteConverter.ToUInt16(buffer, 0);
				int messagelen = 4+16+len;
				m_SocketHandler.Read(messagelen, new SocketHandler.delRead(HandlePackChatMessageQuote), PrimaryForm);
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackChatMessageQuote(byte[] buffer)
		{
			try
			{
				int index = 0;
				int fromid = ByteConverter.ToInt32(buffer, ref index);
				Guid chatid = ByteConverter.ToGUID(buffer, ref index);
				
				string quote = Encoding.ASCII.GetString(buffer, index, buffer.Length-index);
				PrimaryForm.PostChatMessage(chatid, fromid, quote);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackUserEnterChat(byte[] buffer)
		{
			try
			{
				int index = 0;
				Guid chatid = ByteConverter.ToGUID(buffer, ref index);
				int fromid = ByteConverter.ToInt32(buffer, ref index);
				string name = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXNAME).TrimEnd('\0');
				index += CONSTANTS.MAXNAME;
				string clan = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXCLAN).TrimEnd('\0');
				PrimaryForm.UserChatAdd(chatid, fromid, name, clan, true);
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackUserLeaveChat(byte[] buffer)
		{
			int index = 0;
			Guid chatid = ByteConverter.ToGUID(buffer, ref index);
			int userid = ByteConverter.ToInt32(buffer, ref index);
			PrimaryForm.UserChatRemove(chatid, userid);
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackBootFromChat(byte[] buffer)
		{
			int charid = ByteConverter.ToInt32(buffer, 0);
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackCreateChat(byte[] buffer)
		{
			int index = 0;
			Guid chatid = ByteConverter.ToGUID(buffer, ref index);
			string chatname = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXCHATNAME).TrimEnd('\0');
			PrimaryForm.AddChat(chatname, chatid, 1);
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackFailedCreateChat()
		{
			MessageBox.Show("Failed to create chatroom", "Error");
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackChatDestroyed(byte[] buffer)
		{
			int index = 0;
			Guid chatid = ByteConverter.ToGUID(buffer, ref index);
			PrimaryForm.RemoveChat(chatid);
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackBattleCreateSuccess(byte[] buffer)
		{
			uint id = ByteConverter.ToUInt32(buffer, 0);
			ushort port = ByteConverter.ToUInt16(buffer, 4);
			string mapname = Encoding.ASCII.GetString(buffer, 6, CONSTANTS.MAXMAPNAME).TrimEnd("\0".ToCharArray());
			byte password = buffer[6+CONSTANTS.MAXMAPNAME];
			byte maxplayers = buffer[7+CONSTANTS.MAXMAPNAME];
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			PrimaryForm.HostBattle(id, port, mapname, password, maxplayers);
		}

		private void HandlePackBattleCreateFailed()
		{
			Global.battleRoom = null;
			MessageBox.Show("Failed to create battle");
			m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
		}

		private void HandlePackBattleProgress(byte[] buffer )
		{
			try
			{
				uint id = ByteConverter.ToUInt32(buffer, 0);
				byte state = buffer[4];
				PrimaryForm.BattleProgressMessage(id, state);
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackCreateBattle(byte[] buffer)
		{
			try
			{
				BattleHostInfo bhi = new BattleHostInfo();
				bhi.id = ByteConverter.ToUInt32(buffer, 0);
				bhi.battlename = Encoding.ASCII.GetString(buffer, 4, CONSTANTS.MAXBATTLENAME).TrimEnd("\0".ToCharArray());
				bhi.clan = Encoding.ASCII.GetString(buffer, 4+CONSTANTS.MAXBATTLENAME, CONSTANTS.MAXCLAN).TrimEnd("\0".ToCharArray());
				bhi.hostername = Encoding.ASCII.GetString(buffer, 4+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN, CONSTANTS.MAXNAME).TrimEnd("\0".ToCharArray());
				bhi.port = ByteConverter.ToUInt16(buffer, 4+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME);
				bhi.numplayers = buffer[6+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME];
				bhi.interfaceaddress = ByteConverter.ToUInt32(buffer, 7+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME);
				bhi.mapname = Encoding.ASCII.GetString(buffer, 11+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME, CONSTANTS.MAXMAPNAME).TrimEnd("\0".ToCharArray());
				bhi.currplayers = 1;
				if ( buffer[11+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME+CONSTANTS.MAXMAPNAME] == 0 )
					bhi.password = false;
				else
					bhi.password = true;
				bhi.inProgress = false;
				PrimaryForm.AddBattle(bhi);
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackDestroyBattle(byte[] buffer)
		{
			try
			{
				uint id = ByteConverter.ToUInt32(buffer, 0);
				PrimaryForm.DestroyBattleMessage(id);
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackBattleList(byte[] buffer)
		{	
			try
			{
				int sizeMessage = 14+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME+CONSTANTS.MAXMAPNAME;
				int numBattles = ByteConverter.ToInt32(buffer, 0);
				sizeMessage *= numBattles;
				if ( sizeMessage == 0 )
				{
					// continue the message loop
					m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
					return;
				}
				socketHandler.Read(sizeMessage, new SocketHandler.delRead(HandlePackBattleListItems), PrimaryForm);
				// continue the message loop
				m_SocketHandler.Read(1, new SocketHandler.delRead(MessageHandler), PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackBattleListItems(byte[] buffer)
		{
			try
			{
				int sizeItem = 13+CONSTANTS.MAXBATTLENAME+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME+CONSTANTS.MAXMAPNAME;
				int numBattles = buffer.Length / sizeItem;
				int index = 0;
				for ( int i = 0; i < numBattles; i++ )
				{
					BattleHostInfo bhi = new BattleHostInfo();
					bhi.id = ByteConverter.ToUInt32(buffer, index);
					index += 4;
					bhi.battlename = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXBATTLENAME).TrimEnd("\0".ToCharArray());
					index += CONSTANTS.MAXBATTLENAME;
					bhi.clan = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXCLAN).TrimEnd("\0".ToCharArray());
					index += CONSTANTS.MAXCLAN;
					bhi.hostername = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXNAME).TrimEnd("\0".ToCharArray());
					index += CONSTANTS.MAXNAME;
					bhi.port = ByteConverter.ToUInt16(buffer, index);
					index += 2;
					bhi.numplayers = buffer[index];
					index++;
					bhi.interfaceaddress = ByteConverter.ToUInt32(buffer, index);
					index += 4;
					bhi.mapname = Encoding.ASCII.GetString(buffer, index, CONSTANTS.MAXMAPNAME).TrimEnd("\0".ToCharArray());
					index += CONSTANTS.MAXMAPNAME;
					if ( buffer[index] == 0 )
						bhi.password = false;
					else
						bhi.password = true;
					index++;
					bhi.currplayers = buffer[index++];
					bhi.inProgress = buffer[index++] == 0 ? false : true;
					PrimaryForm.AddBattle(bhi);
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void Disconnect()
		{
			if ( s != null && s.Connected )
			{
				s.Shutdown(SocketShutdown.Both);
				s.Close();
			}
		}

		// the following all uses blocking I/O.  this is useful pretty much only when logging in
		#region BlockingI/O
		public bool Send(byte[] data, int offset, int amount)
		{
			if ( !s.Connected || (offset + amount) > data.Length )
				return false;
			int sent = 0;
			do
			{
				sent += s.Send(data, offset+sent, amount-sent, SocketFlags.None);
			}
			while ( sent < amount );

			return true;
		}

		public bool Send(byte[] data, int amount )
		{
			if ( !s.Connected || amount > data.Length )
				return false;
			int sent = 0;
			do
			{
				sent += s.Send(data, sent, amount-sent, SocketFlags.None);
			}
			while ( sent < amount );

			return true;
		}

		public bool Send(byte[] data)
		{
			if ( !s.Connected )
				return false;

			int sent = 0;
			do
			{
				sent += s.Send(data, sent, data.Length-sent, SocketFlags.None);
			}
			while ( sent < data.Length );

			return true;
		}

		public bool Send(int data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[4];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );

			return true;
		}

		public bool Send(Guid data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[16];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );

			return true;
		}

		public bool Send(uint data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[4];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(short data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[2];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(ushort data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[2];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(long data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[8];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(ulong data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[8];
			ByteConverter.ToNetworkBytes(data,info,0);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(string data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = Encoding.ASCII.GetBytes(data);
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Send(byte data)
		{
			if ( !s.Connected )
				return false;
			byte[] info = new byte[1];
			info[0] = data;
			
			int sent = 0;
			do
			{
				sent += s.Send(info, sent, info.Length-sent, SocketFlags.None);
			}
			while ( sent < info.Length );
			return true;
		}

		public bool Read(byte[] dest, int offset, int amount)
		{
			if ( !s.Connected )
				return false;
			int received = 0;
			do
			{
				received += s.Receive(dest, received+offset, amount-received, SocketFlags.None);
			}
			while ( received < amount );

			return true;
		}

		public bool Read(byte[] dest)
		{
			if ( !s.Connected )
				return false;
			int received = 0;
			do
			{
				received += s.Receive(dest, received, dest.Length-received, SocketFlags.None);
			}
			while ( received < dest.Length );
	
			return true;
		}
		#endregion
	}
}
