#include "stdafx.h"
#include ".\grounddecalhandler.h"
#include <algorithm>
#include "bitmap.h"
#include "mygl.h"
#include "vertexarray.h"
#include "unit.h"
#include "ground.h"
#include "unitdef.h"
#include "infoconsole.h"
#include "reghandler.h"
#include <cctype>
#include "camera.h"

CGroundDecalHandler* groundDecals=0;
using namespace std;

CGroundDecalHandler::CGroundDecalHandler(void)
{
	decalLevel=regHandler.GetInt("GroundDecals",1);
/*
	unsigned char buf[128*128*4];

	for(int y=0;y<128;++y){
		for(int x=0;x<128;++x){
			buf[(y*128+x)*4+0]=180;//60+gu->usRandFloat()*20;
			buf[(y*128+x)*4+1]=0;//40+gu->usRandFloat()*10;
			buf[(y*128+x)*4+2]=255;//20+gu->usRandFloat()*5;
			buf[(y*128+x)*4+3]=0;
		}
	}
	for(int y=0;y<20;++y){
		for(int x=0;x<128;++x){
			buf[(y*128+x)*4+1]=255*min(1.0,max(0.4,sin(x*PI/16)*0.5+0.15))*(y==0?0:1);
			buf[((y+108)*128+x)*4+1]=255*min(1.0,max(0.4,sin(x*PI/16)*0.5+0.15))*(y==19?0:1);
		}
	}
	CBitmap bm(buf,128,128);
	bm.Save("StdTank.bmp");
	tankTex=bm.CreateTexture(true);*/
}

CGroundDecalHandler::~CGroundDecalHandler(void)
{
	for(std::vector<TrackType*>::iterator tti=trackTypes.begin();tti!=trackTypes.end();++tti){
		for(set<UnitTrackStruct*>::iterator ti=(*tti)->tracks.begin();ti!=(*tti)->tracks.end();++ti){
			delete *ti;
		}
		glDeleteTextures (1, &(*tti)->texture);
		delete *tti;
	}
}

void CGroundDecalHandler::Draw(void)
{
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glEnable(GL_ALPHA_TEST);
	glAlphaFunc(GL_GREATER,0.01);
	glDepthMask(0);
	glPolygonOffset(-10,-10);
	glEnable(GL_POLYGON_OFFSET_FILL);

	CVertexArray* va=GetVertexArray();
	va->Initialize();

	unsigned char color[4];
	color[0]=255;
	color[1]=255;
	color[2]=255;
	color[3]=255;
	unsigned char color2[4];
	color2[0]=255;
	color2[1]=255;
	color2[2]=255;
	color2[3]=255;

	for(std::vector<TrackType*>::iterator tti=trackTypes.begin();tti!=trackTypes.end();++tti){
		if(!(*tti)->tracks.empty())
			glBindTexture(GL_TEXTURE_2D,(*tti)->texture);
		for(set<UnitTrackStruct*>::iterator ti=(*tti)->tracks.begin();ti!=(*tti)->tracks.end();){
			UnitTrackStruct* track=*ti;
			while(!track->parts.empty() && track->parts.front().creationTime<gs->frameNum-track->lifeTime){
				track->parts.pop_front();
			}
			if(track->parts.empty()){
				if(track->owner)
					track->owner->myTrack=0;
				delete track;
				ti=(*tti)->tracks.erase(ti);
				continue;
			}
			if(camera->InView((track->parts.front().pos1+track->parts.back().pos1)*0.5,track->parts.front().pos1.distance(track->parts.back().pos1)+500)){
				list<TrackPart>::iterator ppi=track->parts.begin();
				color2[3]=track->trackAlpha-(gs->frameNum-ppi->creationTime)*track->alphaFalloff;

				for(list<TrackPart>::iterator pi=++track->parts.begin();pi!=track->parts.end();++pi){
					color[3]=track->trackAlpha-(gs->frameNum-ppi->creationTime)*track->alphaFalloff;
					if(pi->connected){
						va->AddVertexTC(ppi->pos1,ppi->texPos,0,color2);
						va->AddVertexTC(ppi->pos2,ppi->texPos,1,color2);
						va->AddVertexTC(pi->pos2,pi->texPos,1,color);
						va->AddVertexTC(pi->pos1,pi->texPos,0,color);
					}
					color2[3]=color[3];
					ppi=pi;
				}
			}
			++ti;
		}
		va->DrawArrayTC(GL_QUADS);
	}

	glDisable(GL_POLYGON_OFFSET_FILL);
//	glDisable(GL_ALPHA_TEST);
	glDepthMask(1);
	glDisable(GL_BLEND);
}

void CGroundDecalHandler::Update(void)
{
}

void CGroundDecalHandler::UnitMoved(CUnit* unit)
{
	TrackType* type=trackTypes[unit->unitDef->trackType];
	if(!unit->myTrack){
		UnitTrackStruct* ts=new UnitTrackStruct;
		ts->owner=unit;
		ts->lifeTime=30*decalLevel*unit->unitDef->trackStrength;
		ts->trackAlpha=unit->unitDef->trackStrength*25;
		ts->alphaFalloff=float(ts->trackAlpha)/float(ts->lifeTime);
		type->tracks.insert(ts);
		unit->myTrack=ts;
	}
	float3 pos=unit->pos+unit->frontdir*unit->unitDef->trackOffset;

	TrackPart tp;
	tp.pos1=pos+unit->rightdir*unit->unitDef->trackWidth*0.5;
	tp.pos1.y=ground->GetHeight2(tp.pos1.x,tp.pos1.z);
	tp.pos2=pos-unit->rightdir*unit->unitDef->trackWidth*0.5;
	tp.pos2.y=ground->GetHeight2(tp.pos2.x,tp.pos2.z);
	tp.creationTime=gs->frameNum;
	if(unit->myTrack->parts.empty()){
		tp.texPos=0;
		tp.connected=false;
	} else {
		tp.texPos=unit->myTrack->parts.back().texPos+tp.pos1.distance(unit->myTrack->parts.back().pos1)/unit->unitDef->trackWidth*unit->unitDef->trackStretch;
		tp.connected=unit->myTrack->parts.back().creationTime==gs->frameNum-8;
	}

	unit->myTrack->parts.push_back(tp);
}

void CGroundDecalHandler::RemoveUnit(CUnit* unit)
{
	if(unit->myTrack){
		unit->myTrack->owner=0;
		unit->myTrack=0;
	}
}

int CGroundDecalHandler::GetTrackType(std::string name)
{
	std::transform(name.begin(), name.end(), name.begin(), (int (*)(int))std::tolower);	

	int a=0;
	for(std::vector<TrackType*>::iterator ti=trackTypes.begin();ti!=trackTypes.end();++ti){
		if((*ti)->name==name){
			return a;
		}
		++a;
	}
	TrackType* tt=new TrackType;
	tt->name=name;
	tt->texture=LoadTexture(name);
	trackTypes.push_back(tt);
	return trackTypes.size()-1;
}

unsigned int CGroundDecalHandler::LoadTexture(std::string name)
{
	if(name.find_first_of('.')==string::npos)
		name+=".bmp";
	CBitmap bm(string(".\\bitmaps\\tracks\\")+name);
	for(int y=0;y<bm.ysize;++y){
		for(int x=0;x<bm.xsize;++x){
			bm.mem[(y*bm.xsize+x)*4+3]=bm.mem[(y*bm.xsize+x)*4+1];
			int brighness=bm.mem[(y*bm.xsize+x)*4+0];
			bm.mem[(y*bm.xsize+x)*4+0]=(brighness*90)/255;
			bm.mem[(y*bm.xsize+x)*4+1]=(brighness*60)/255;
			bm.mem[(y*bm.xsize+x)*4+2]=(brighness*30)/255;
		}
	}
	
	return bm.CreateTexture(true);
}
